from database_mysql_local.connector import Connector
from logger_local.LoggerLocal import Logger
from variable_local.variable import VariablesLocal

from .Constants import DIALOG_WORKFLOW_CODE_LOGGER_OBJECT
from .utils import get_curr_state

logger = Logger.create_logger(object=DIALOG_WORKFLOW_CODE_LOGGER_OBJECT)


class ProfileContext(object):
    # TODO We should consider to take ProfileContact to a separate package or merge it with UserContext
    def __init__(self, profile_id):
        self.dict = {}
        self.profile_id = profile_id
        self.chosen_poll_options = {}
        self.curr_state_id = get_curr_state(self.profile_id)
        self.variables = VariablesLocal()
        self.groups = []

    def get_variable_value_by_variable_id(self, variable_id: int) -> str:
        logger.start(object={'variable_id': variable_id})
        variable_value = self.dict[variable_id]
        logger.end(object={'variable_value': variable_value})
        return variable_value

    def save_chosen_options(self, question_asked: str, variable_id: int, chosen_numbers_list: list,
                            list_of_options: list):
        """Saves the options chosen by the user in the multi_choice_poll action in a dict with the question as the key
            and a list of the options chosen as the value i.e: {<question asked> : [<chosen option 1>, <chosen option 2>, ...]}
            Also saves the chosen options in the database."""
        logger.start(object={'question_asked': question_asked, 'variable_id': variable_id,
                             'chosen_numbers_list': chosen_numbers_list, 'list_of_options': list_of_options})
        self.chosen_poll_options[question_asked] = [
            list_of_options[chosen_option - 1] for chosen_option in chosen_numbers_list]
        variable_value_to_insert = question_asked + " "
        for chosen_option in self.chosen_poll_options[question_asked]:
            variable_value_to_insert = variable_value_to_insert + \
                                       str(chosen_option) + ", "
        self.variables.set_variable_value_by_variable_id(
            variable_id, variable_value_to_insert, self.profile_id, self.curr_state_id)
        logger.end()

    def get_variable_value_by_variable_name(self, variable_name: str) -> str:
        logger.start(object={'variable_name': variable_name})
        variable_value = self.get_variable_value_by_variable_id(
            self.variables.get_variable_id_by_variable_name(variable_name))
        logger.end(object={'variable_value': variable_value})
        return variable_value

    # TODO Should be moved to variable-value-local-python-package
    def set(self, variable_id: int, variable_value: str):
        logger.start(object={'variable_id': variable_id, 'variable_value': variable_value})
        self.dict[variable_id] = variable_value
        connection = Connector.connect('logger')
        cursor = connection.cursor(dictionary=True, buffered=True)
        # cursor.execute("""USE logger""")
        # TODO Can we replace it with GenericCRUD
        cursor.execute("""INSERT INTO logger
                        (profile_id, state_id, variable_id, variable_value_new) 
                        VALUES (%s,%s,%s,%s)""",
                       (self.profile_id, self.curr_state_id, variable_id, variable_value))
        connection.commit()
        logger.end()


class ProfilesDict(object):
    def __init__(self):
        self.profiles_dict = {}

    def add(self, profile: ProfileContext):
        logger.start(object={'profile': profile})
        # TODO Should we also keep in the database?
        self.profiles_dict[profile.profile_id] = profile
        logger.end()

    def get(self, profile_id: int) -> ProfileContext or None:
        logger.start(object={'profile_id': profile_id})
        if profile_id not in self.profiles_dict:
            logger.end()
            return None
        else:
            profile_dict = self.profiles_dict[profile_id]
            logger.end(object={'profile_dict': profile_dict})
            return profile_dict
        # return None if profile_id not in self.profiles_dict else self.profiles_dict[profile_id]


class DialogWorkflowRecord(object):
    def __init__(self, record: dict) -> None:
        self.curr_state_id: int = record.get("state_id")
        self.parent_state_id: int = record.get("parent_state_id")
        self.workflow_action_id: int = record.get("workflow_action_id")
        self.lang_code = record.get("lang_code")
        self.parameter1 = record.get("parameter1")
        self.variable1_id: int = record.get("variable1_id")
        self.result_logical = record.get("result_logical")
        self.result_figure_min: float = record.get("result_figure_min")
        self.result_figure_max: float = record.get("result_figure_max")
        self.next_state_id: int = record.get("next_state_id")
        self.no_feedback_milliseconds: float = record.get("no_feedback_milliseconds")
        self.next_state_id_if_there_is_no_feedback: int = record.get("next_state_id_if_there_is_no_feedback")



from enum import Enum

from logger_local.LoggerComponentEnum import LoggerComponentEnum
from python_sdk_remote.utilities import our_get_env


# TODO Let's use SQL2Code

# TODO Move to a seperate file
# TODO Use Sql2Code
class WorkflowActionEnum(Enum):
    PRESENT_AND_CHOOSE_SCRIPT = 0
    LABEL_ACTION = 1
    TEXT_MESSAGE_ACTION = 2
    QUESTION_ACTION = 3
    JUMP_ACTION = 4
    SEND_REST_API_ACTION = 5
    ASSIGN_VARIABLE_ACTION = 6
    INCREMENT_VARIABLE_ACTION = 7
    DECREMENT_VARIABLE_ACTION = 8
    CONDITION_ACTION = 9
    MENU_ACTION = 10
    AGE_DETECTION = 11
    MULTI_CHOICE_POLL = 12
    PRESENT_CHILD_GROUPS_NAMES_BY_ID = 13
    PRESENT_GROUPS_WITH_CERTAIN_TEXT = 14
    INSERT_MISSING_DATA = 15
    PRESENT_FORM = 16


# TODO Move to a seperate file
# TODO Use Sql2Code
# TODO Can we load it one time from database to memcache
VARIABLE_NAMES_DICT = {1: "Person Id", 2: "User Id", 3: "Profile Id", 4: "Lang Code",
                       5: "Name Prefix", 6: "First Name", 7: "Middle Name",
                       8: "Last Name", 9: "Name Suffix", 10: "Full Name",
                       11: "Country", 12: "State", 13: "County", 14: "City",
                       15: "Neighborhood", 16: "Street", 17: "House", 18: "Suite/Apartment",
                       19: "Zip Code", 20: "Post Result", 21: "Age", 22: "Result"}


class CommunicationTypeEnum(Enum):
    CONSOLE = 1
    WEBSOCKET = 2
    MESSAGE = 3


# TODO: We should align the terminology between dialog-workflow and gender-detection (INTERFACE_MODE= BATCH, INTERACTIVE)
DEFAULT_COMMUNICATION_TYPE = CommunicationTypeEnum.WEBSOCKET.value
COMMUNICATION_TYPE = CommunicationTypeEnum(int(our_get_env("COMMUNICATION_TYPE",
                                                           default=str(DEFAULT_COMMUNICATION_TYPE))))

DIALOG_WORKFLOW_PYTHON_PACKAGE_COMPONENT_ID = 166
DIALOG_WORKFLOW_PYTHON_PACKAGE_COMPONENT_NAME = "dialog_workflow-python-package"

DIALOG_WORKFLOW_CODE_LOGGER_OBJECT = {
    'component_id': DIALOG_WORKFLOW_PYTHON_PACKAGE_COMPONENT_ID,
    'component_name': DIALOG_WORKFLOW_PYTHON_PACKAGE_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': 'idan.a@circ.zone and guy.n@circ.zone'
}

LOGGER_TEST_OBJECT = DIALOG_WORKFLOW_CODE_LOGGER_OBJECT.copy()
LOGGER_TEST_OBJECT['component_category'] = LoggerComponentEnum.ComponentCategory.Unit_Test.value



import random

from database_mysql_local.connector import Connector
from logger_local.LoggerLocal import Logger
from message_local.CompoundMessage import CompoundMessage
from user_context_remote.user_context import UserContext
from variable_local.variable import VariablesLocal
from language_remote.lang_code import LangCode

from .Constants import DIALOG_WORKFLOW_CODE_LOGGER_OBJECT
from .ProfileContext import DialogWorkflowRecord
from .action import Action
from .utils import get_curr_state, update_profile_curr_state_in_db

user = UserContext()
logger = Logger.create_logger(object=DIALOG_WORKFLOW_CODE_LOGGER_OBJECT)


# Get all potential records in a specific state and choose randomly one of them
# TODO Please use MessagesLocal to store the message(s) recieved and anwers
# TODO Please use the function defined in avatar similar to get_dialog_workflow_avatar_profile_id( profile_id1, prompt, group_id, profile_id2...) using avatar_table and avtar_group_table
def get_dialog_workflow_record(profile_curr_state: int, language: LangCode):
    logger.start(
        object={'profile_curr_state': profile_curr_state, 'language': language})
    # TODO Change * to the fields required
    connection = Connector.connect('dialog_workflow')
    cursor = connection.cursor(dictionary=True, buffered=True)
    cursor.execute("""SELECT * FROM dialog_workflow_state_view WHERE state_id = %s AND lang_code = %s""",
                   (profile_curr_state, language.value))
    optional_records = cursor.fetchall()
    if not optional_records:
        error_message = f"No records found for state_id: {profile_curr_state} and language: {language}"
        logger.error(error_message)
        logger.end()
        raise Exception(error_message)
    random_index = random.randint(0, len(optional_records) - 1)
    dialog_workflow_record = DialogWorkflowRecord(optional_records[random_index])
    logger.end(object={'dialog_workflow_record': str(dialog_workflow_record)})
    return dialog_workflow_record


def post_message(profile_id: int, incoming_message: str):
    """This function is supposed to serve as a POST request later on using REST API.
    It runs until needing input from the user, which it then sends a json to the user with the message and exits
    PARAMS: 
        1. profile_id: the profile id that sent the request
        2. incoming_message: the message he sent"""
    # TODO: remove profile_id an use user context
    logger.start(object={'profile_id': profile_id,
                         'incoming_message': incoming_message})
    # TODO Save the message using MessagesLocal (from DIALOG_WORKFLOW_PROFILE_ID, to UserContext.getEffectiveProfileId, ...)
    variables = VariablesLocal()
    profile_curr_state = get_curr_state(profile_id)
    language = user.get_effective_profile_preferred_lang_code()
    got_response = incoming_message.strip() != ""  # This variable indicates if we must act now as we got a response from the user or as if we should send one to him
    init_action = Action(incoming_message, profile_id,
                         language, profile_curr_state, variables)
    while True:
        dialog_workflow_record = get_dialog_workflow_record(
            init_action.profile_curr_state, language)
        is_state_changed, outgoing_message = init_action.act(
            dialog_workflow_record, got_response)
        # TODO Save the message using MessagesLocal (from DIALOG_WORKFLOW_PROFILE_ID, to UserContext.getEffectiveProfileId, ...)
        if outgoing_message is not None:
            if not isinstance(outgoing_message, list):
                outgoing_compound_message = CompoundMessage(body=outgoing_message).get_message_fields()
            else:
                outgoing_compound_message = outgoing_message
            logger.end(object={'outgoing_message': str(outgoing_compound_message)})
            return outgoing_compound_message
        init_action.profile_curr_state = dialog_workflow_record.next_state_id if is_state_changed == False else init_action.profile_curr_state
        update_profile_curr_state_in_db(
            init_action.profile_curr_state, profile_id)
        got_response = False
